#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lbfileio.h"
#include "bulcommn.h"
#include "keycodes.h"

const int SAVES_COUNT=8;

// Data types

typedef struct {
        unsigned char enabled;
        unsigned char zero;
        char name[15];
       } CAT_ENTRY;

typedef struct {
        char comment[25];
        unsigned char filesize[4];
        unsigned char headersize[4];
       } SAVEHEADER;

enum gameType {gtUnknown = 0x000,gtKeeper = 0x001,gtDeeper = 0x002};

enum gameVersion
    {gvDKUnkn = 0x000,  gvDKBeta = 0x001, gvDKFinal = 0x002, gvDKRetail = 0x003,
     gvDDUnkn = 0x004,  gvDDBeta = 0x005, gvDDFinal = 0x006, gvDDRetail = 0x007 };
const unsigned int save_gold_whole_offs[]=
    {0x143FD6,          0x143FD6,         0x143FD6,          0x143FD6,
     0x143FD7,          0x143FD7,         0x143FD7,          0x143FD7};
const unsigned int save_gold_free_offs[]=
    {0x143FDA,          0x143FDA,         0x143FDA,          0x143FDA,
     0x143FDB,          0x143FDB,         0x143FDB,          0x143FDB};


char *getCatFileName(int game_type)
{
  switch (game_type)
  {
  case gtKeeper:
      return ".\\save\\save.cat";
  case gtDeeper:
      return ".\\save\\dd1save.cat";
  }
  return NULL;
}

void getSaveFileName(char *dest, int game_type,int num)
{
  switch (game_type)
  {
  case gtKeeper:
      sprintf (dest,".\\save\\game%04u.sav",num);
      break;
  case gtDeeper:
      sprintf (dest,".\\save\\dd1g%04u.sav",num);
      break;
  }
}

int bit1Count(unsigned int x)
{
    int count;
    for (count=0; x != 0; x>>=1)
    if (x & 01) count++;
    return count;
}

int getUsrUValue(int minVal,int maxVal,int *newval)
{
         (*newval)=minVal-1;
         if ((*newval)>-1) (*newval)=-1;
         printf("Enter new value (%d..%d):",minVal,maxVal);
         char buf[255];
         scanf("%s",buf);
         sscanf(buf,"%i",newval);
         if (((*newval)>=minVal) && ((*newval)<=maxVal))
         {
            printf("\nValue %d accepted.\n",*newval);
            return 1;
         }
         else
            printf("\nValue %d out of range, ignored.\n",*newval);
         return 0;
}

int getUsr8bitUValueTo(int minVal,int maxVal,unsigned char *dest_ptr)
{
    int newval;
    if (!getUsrUValue(minVal,maxVal,&newval))
        return 0;
    (*dest_ptr)=(unsigned char)(newval&255);
}

int saveStructureToFile(char *samplename,char *src_ptr,unsigned long length)
{
         printf("\nEnter destination file name, ie \"%s\":",samplename);
         char newname[255];
         int i;
         for (i=0;i<255;i++)
             newname[i]=0;
         scanf("%s",newname);
         if ((strlen(newname)<1)||(newname[0]<'0')) strcpy(newname,samplename);
         FILE *expfp=fopen(newname,"wb");
         if (expfp==NULL)
         {
            printf("\nOperation skipped - file creation error.\n");
            return 1;
         }
         fwrite (src_ptr, length, 1, expfp);
         fclose(expfp);         
         return 0;
}

int loadStructureFromFile(char *samplename,char *dest_ptr,unsigned long length)
{
         printf("\nEnter source file name, ie \"%s\":",samplename);
         char newname[255];
         int i;
         for (i=0;i<255;i++)
             newname[i]=0;
         scanf("%s",newname);
         if ((strlen(newname)<1)||(newname[0]<'0')) strcpy(newname,samplename);
         FILE *impfp=fopen(newname,"rb");
         if (impfp==NULL)
         {
            printf("\nOperation skipped - file opening error.\n");
            return 1;
         }
         fread (dest_ptr, length, 1, impfp);
         fclose(impfp);
         return 0;
}

void getSavesCatalog(int game_type,CAT_ENTRY save_list[2*SAVES_COUNT],int first_idx)
//Gets text comments inside saves, returns number of saves
{
  int num;
  for (num=first_idx;num<(first_idx+SAVES_COUNT+1);num++)
  {
      save_list[num].enabled=0;
      save_list[num].zero=0;
      sprintf(save_list[num].name,"");
  }
  FILE *catfp = fopen (getCatFileName(game_type), "rb");
  if (!catfp) return;
  for (num=first_idx;num<(first_idx+SAVES_COUNT+1);num++)
  {
      fread (&(save_list[num]), 17, 1, catfp);
  }
  fclose(catfp);
}

int editGold(char *save_data,int game_version)
{
    printf("Treasury refiller - enter new gold amount.\n");
    unsigned int gold_whole=read_int32_le_buf((unsigned char *)&(save_data[save_gold_whole_offs[game_version]]));
    unsigned int gold_free=read_int32_le_buf((unsigned char *)&(save_data[save_gold_free_offs[game_version]]));
    unsigned int gold_min=0;
    if (gold_whole>gold_free) gold_min=gold_whole-gold_free;
    int newval;
    if (!getUsrUValue(gold_min,2147483647,&newval))
        return 0;
    gold_whole=newval;
    gold_free=newval-gold_min;
    write_int32_le_buf((unsigned char *)&(save_data[save_gold_whole_offs[game_version]]),gold_whole);
    write_int32_le_buf((unsigned char *)&(save_data[save_gold_free_offs[game_version]]), gold_free);
    return 0;
}

int editSavegame(int game_type,int save_num)
{
    printf("Reading savegame...");
    FILE *savefp;
    char savefname[255];
    getSaveFileName(savefname,game_type,save_num);
    savefp = fopen (savefname, "rb");
    if (!savefp)
    {
        printf("\nCannot open savegame file!\n");
        return 1;
    }
    int savfile_rsize=file_length_opened (savefp);
    char *save_data=malloc(savfile_rsize);
    fread (save_data, savfile_rsize, 1, savefp);
    printf("done.\n");
    fclose(savefp);
    printf("Detecting save version...");
    int game_version;
    if (game_type==gtKeeper)
    {
      if (savfile_rsize==1382437)
      {
        game_version=gvDKRetail;
        printf("DK Retail");
      } else
      if (savfile_rsize==1386066)
      {
        game_version=gvDKBeta;
        printf("DK Beta");
      } else
      {
        game_version=gvDKUnkn;
        printf("unknown DK");
      }
    } else
    if (game_type==gtDeeper)
    {
      if (savfile_rsize==1382458)
      {
        game_version=gvDDRetail;
        printf("DD Retail");
      } else
      {
        game_version=gvDDUnkn;
        printf("unknown DD");
      }
    } else
    {
      game_version=gvDKUnkn;
      printf("unknown DK");
    }
    printf(" identified.\n");
  if ((game_version==gvDKUnkn)||(game_version==gvDDUnkn))
      printf("Warning: Unknown version of savegame! Be cautious!\n");
    unsigned int savfile_ssize=0;//read_int32_le_buf((unsigned char *)&(save_head.filesize));
    printf("Save file size: in save %d, real %d\n",savfile_ssize,savfile_rsize);
  while (1)
  {
    printf("\nSELECT SAVE GAME OPERATION:\n");
    unsigned int val_money=read_int32_le_buf((unsigned char *)&(save_data[save_gold_whole_offs[game_version]]));
    printf("%3d - %s: %10u\n",1,"Gold amount",val_money);
    printf("%3d - %s\n",9,"Save changes into savegame");
    printf("Esc - %s\n","Exit");
    printf("\nEnter function number:");
    char usr_answer;
    do
    {
        usr_answer=getch();
    }
    while ((usr_answer!=VK_ESCAPE)&&(usr_answer<'1')&&(usr_answer>'9'));
    printf("%c\n",usr_answer);
    int result=0;
    if (usr_answer==VK_ESCAPE)
    {
        printf("\n\nExiting - have a nice play!\n");
        result=1;
    }
    switch (usr_answer)
    {
    case '1':
         editGold(save_data,game_version);
         break;
/*
    case '1':
         printf("\nEnter new savegame comment(max. %u chars):\n",sizeof(save_head.comment)-1);
         char newname[255];
         int i;
         for (i=0;i<255;i++)
             newname[i]=0;
         scanf("%s",newname);
         if ((strlen(newname)<1)||(newname[0]<'!'))
           printf("\nTyped string cannot name a savegame! Ignored.\n");
         else
         {
          strncpy(save_head.comment,newname,sizeof(save_head.comment)-1);
           printf("\nComment updated.\n");
         }
         result=0;
         break;
    case '3':
         editCyros(save_data);
         break;
    case '4':
         editResearch(save_data);
         break;
    case '5':
        printf("\nSorry - unsupported yet.\n");
        system("PAUSE");
        result=0;
        break;
    case '6':
         editVarious(save_data);
         break;
    //Import/export
    case '7':
      {
         char defaultname[255];
         sprintf (defaultname,".\\sw_dec%d.dat",save_num);
         result=loadStructureFromFile(defaultname,(char *)save_data, savfile_ssize);
         break;
      }
    case '8':
      {
         char defaultname[255];
         sprintf (defaultname,".\\sw_dec%d.dat",save_num);
         result=saveStructureToFile(defaultname,(char *)save_data, savfile_ssize);
         break;
      }
*/
    case '9':
        printf("\nWriting savegame...");
        savefp = fopen (savefname, "wb");
        fwrite (save_data, savfile_rsize, 1, savefp);
        printf("done.\n");
        fclose(savefp);
        printf("Savegame file has been rewritten.\n");
        result=0;
        break;
    default:
         //printf("\nUnknown function. Select one of known functions!\n");
         break;
    }
    if (result!=0)
    {
      free(save_data);
      return result;
    }
  }
}

int main(int argc, char *argv[])
{
    printf("\nBullfrog Engine Savegame editor for Dungeon Keeper 0.01");
    printf("\n    written by Tomasz Lis, 2000-2007");
    printf("\n-------------------------------\n");
    printf("Listing savegames...");
    CAT_ENTRY save_list[2*SAVES_COUNT];
    getSavesCatalog(gtKeeper,save_list,0);
    getSavesCatalog(gtDeeper,save_list,SAVES_COUNT);
    printf("done.\n");
    printf("Saved games list:\n");
    printf("     %-20s     %-20s\n","Dungeon Keeper:","Deeper Dungeons:");
    int num;
    for (num=0;num<SAVES_COUNT;num++)
    {
        char save_name[255];
        sprintf(save_name,"\"%s\"",save_list[num].name);
        if (save_list[num].enabled)
            printf("    %d - %-20s",num+1,save_name);
        sprintf(save_name,"\"%s\"",save_list[SAVES_COUNT+num].name);
        if (save_list[SAVES_COUNT+num].enabled)
            printf("shift+%d - %-20s",num+1,save_name);
        printf("\n");
    }
    printf("    %c - %s\n",'C',"Cheats");
    printf("  Esc - %s\n","Exit");
    printf("Enter savegame number:");
    int save_num;
    char usr_answer;
    do
    {
      usr_answer=toupper(getch());
      if ((usr_answer<'1')||(usr_answer>'8'))
      {
        switch (usr_answer)
        {
        case '!': save_num=SAVES_COUNT+0;break;
        case '@': save_num=SAVES_COUNT+1;break;
        case '#': save_num=SAVES_COUNT+2;break;
        case '$': save_num=SAVES_COUNT+3;break;
        case '%': save_num=SAVES_COUNT+4;break;
        case '^': save_num=SAVES_COUNT+5;break;
        case '&': save_num=SAVES_COUNT+6;break;
        case '*': save_num=SAVES_COUNT+7;break;
        case '(': save_num=SAVES_COUNT+8;break;
        //case ')': save_num=SAVES_COUNT+0;break;
        default:  save_num=-1;break;
        }
      }
      else
          save_num=usr_answer-'1';
      if (save_num>=2*SAVES_COUNT) save_num=-1;
    }
    while ((usr_answer!=VK_ESCAPE)&&(usr_answer!='C')&&((save_num<0)||(save_list[save_num].enabled==0)));
    if (usr_answer==VK_ESCAPE)
    {
        printf("\n\nExiting - have a nice play!\n");
        return 0;
    }
    if (usr_answer=='C')
    {
        printf("\n\nCHEAT:\n");
        printf("\n");
        printf("   (n/a yet)\n");
        printf("Press any key to exit...");
        char x=getch();
        if (x==0) getch();
        printf("\n");
        return 0;
    }
    printf("%d\n\n",save_num+1);
    unsigned int game_type;
    if (save_num<SAVES_COUNT) game_type=gtKeeper; else game_type=gtDeeper;
    return editSavegame(game_type,save_num%SAVES_COUNT);
}

